@@ -54,6 +54,7 @@
 #include <visualization_msgs/MarkerArray.h>
 #include <mcl_3dl_msgs/ResizeParticle.h>
 #include <mcl_3dl_msgs/Status.h>
+#include <mcl_3dl_msgs/LoadPCD.h>
 #include <std_srvs/Trigger.h>
 #include <diagnostic_updater/diagnostic_updater.h>
 
@@ -71,8 +72,6 @@
 #include <pcl/kdtree/kdtree.h>
 #include <pcl/kdtree/kdtree_flann.h>
 
-#include <pcl18_backports/voxel_grid.h>
-
 #include <mcl_3dl/chunked_kdtree.h>
 #include <mcl_3dl/cloud_accum.h>
 #include <mcl_3dl/filter.h>
@@ -115,6 +114,7 @@ protected:
     MyPointRepresentation()
     {
       nr_dimensions_ = 3;
+      trivial_ = true;
     }
 
     virtual void copyToFloatArray(const PointType& p, float* out) const
@@ -136,22 +136,7 @@ protected:
     const ros::Time map_stamp = (msg->header.stamp != ros::Time()) ? msg->header.stamp : ros::Time::now();
     pcl_conversions::toPCL(map_stamp, pc_tmp->header.stamp);
 
-    pc_map_.reset(new pcl::PointCloud<PointType>);
-    pc_map2_.reset();
-    pc_update_.reset();
-    pcl::VoxelGrid18<PointType> ds;
-    ds.setInputCloud(pc_tmp);
-    ds.setLeafSize(params_.map_downsample_x_, params_.map_downsample_y_, params_.map_downsample_z_);
-    ds.filter(*pc_map_);
-    pc_all_accum_.reset(new pcl::PointCloud<PointType>);
-    has_map_ = true;
-
-    accumClear();
-
-    ROS_INFO("map original: %d points", (int)pc_tmp->points.size());
-    ROS_INFO("map reduced: %d points", (int)pc_map_->points.size());
-
-    cbMapUpdateTimer(ros::TimerEvent());
+    loadMapCloud(pc_tmp);
   }
   void cbMapcloudUpdate(const sensor_msgs::PointCloud2::ConstPtr& msg)
   {
@@ -161,7 +146,7 @@ protected:
       return;
 
     pc_update_.reset(new pcl::PointCloud<PointType>);
-    pcl::VoxelGrid18<PointType> ds;
+    pcl::VoxelGrid<PointType> ds;
     ds.setInputCloud(pc_tmp);
     ds.setLeafSize(params_.update_downsample_x_, params_.update_downsample_y_, params_.update_downsample_z_);
     ds.filter(*pc_update_);
@@ -324,6 +309,11 @@ protected:
       return;
     }
 
+    if (pc_accum_header_.empty())
+    {
+      ROS_ERROR("MCL measure function is called without available pointcloud");
+      return;
+    }
     const std_msgs::Header& header = pc_accum_header_.back();
 
     try
@@ -371,7 +361,7 @@ protected:
     const auto ts = boost::chrono::high_resolution_clock::now();
 
     pcl::PointCloud<PointType>::Ptr pc_local_full(new pcl::PointCloud<PointType>);
-    pcl::VoxelGrid18<PointType> ds;
+    pcl::VoxelGrid<PointType> ds;
     ds.setInputCloud(pc_local_accum_);
     ds.setLeafSize(params_.downsample_x_, params_.downsample_y_, params_.downsample_z_);
     ds.filter(*pc_local_full);
@@ -767,8 +757,8 @@ protected:
       *pc_all_accum_ += *pc_particle;
     }
 
-    if ((header.stamp - match_output_last_ > *params_.match_output_interval_ ||
-         header.stamp < match_output_last_ - ros::Duration(1.0)) &&
+    if ((header.stamp > match_output_last_ + *params_.match_output_interval_ ||
+         header.stamp + ros::Duration(1.0) < match_output_last_) &&
         (pub_matched_.getNumSubscribers() > 0 || pub_unmatched_.getNumSubscribers() > 0))
     {
       match_output_last_ = header.stamp;
@@ -1056,7 +1046,7 @@ protected:
     }
     pcl::PointCloud<PointType>::Ptr points(new pcl::PointCloud<PointType>);
 
-    pcl::VoxelGrid18<PointType> ds;
+    pcl::VoxelGrid<PointType> ds;
     ds.setInputCloud(pc_map_);
     ds.setLeafSize(
         params_.global_localization_grid_,
@@ -1065,10 +1055,7 @@ protected:
     ds.filter(*points);
 
     pcl::KdTreeFLANN<PointType>::Ptr kdtree(new pcl::KdTreeFLANN<PointType>);
-    kdtree->setPointRepresentation(
-        boost::dynamic_pointer_cast<
-            pcl::PointRepresentation<PointType>,
-            MyPointRepresentation>(boost::make_shared<MyPointRepresentation>(point_rep_)));
+    kdtree->setPointRepresentation(point_rep_);
     kdtree->setInputCloud(points);
 
     auto pc_filter = [this, kdtree](const PointType& p)
@@ -1176,12 +1163,57 @@ protected:
     pub_status_.publish(status_);
   }
 
+  void loadMapCloud(const pcl::PointCloud<PointType>::Ptr& map_cloud)
+  {
+    pc_map_.reset(new pcl::PointCloud<PointType>);
+    pc_map2_.reset();
+    pc_update_.reset();
+    pcl::VoxelGrid<PointType> ds;
+    ds.setInputCloud(map_cloud);
+    ds.setLeafSize(params_.map_downsample_x_, params_.map_downsample_y_, params_.map_downsample_z_);
+    ds.filter(*pc_map_);
+    pc_all_accum_.reset(new pcl::PointCloud<PointType>);
+    has_map_ = true;
+
+    accumClear();
+    accum_->reset();
+
+    ROS_INFO("map original: %d points", static_cast<int>(map_cloud->points.size()));
+    ROS_INFO("map reduced: %d points", static_cast<int>(pc_map_->points.size()));
+
+    // output map for visualization purposes:
+    cbMapUpdateTimer(ros::TimerEvent());
+  }
+
+  bool cbLoadPCD(mcl_3dl_msgs::LoadPCD::Request& req, mcl_3dl_msgs::LoadPCD::Response& resp)
+  {
+    ROS_INFO("map received");
+
+    pcl::PointCloud<PointType>::Ptr pc_tmp(new pcl::PointCloud<PointType>);
+    if (pcl::io::loadPCDFile<PointType>(req.pcd_path, *pc_tmp) == -1)
+    {
+      ROS_ERROR_STREAM("Couldn't read file " << req.pcd_path);
+      has_map_ = false;
+      resp.success = false;
+      return true;
+    }
+
+    pcl_conversions::toPCL(ros::Time::now(), pc_tmp->header.stamp);
+    pc_tmp->header.frame_id = params_.frame_ids_["map"];
+
+    loadMapCloud(pc_tmp);
+
+    resp.success = true;
+    return true;
+  }
+
 public:
   MCL3dlNode()
     : pnh_("~")
     , tfl_(tfbuf_)
     , cnt_measure_(0)
     , global_localization_fix_cnt_(0)
+    , point_rep_(new MyPointRepresentation)
     , engine_(seed_gen_())
   {
   }
@@ -1197,20 +1229,26 @@ public:
 
     if (!params_.fake_odom_)
     {
+      int odom_queue_size;
+      pnh_.param("odom_queue_size", odom_queue_size, 200);
       sub_odom_ = mcl_3dl_compat::subscribe(
           nh_, "odom",
-          pnh_, "odom", 200, &MCL3dlNode::cbOdom, this);
+          pnh_, "odom", odom_queue_size, &MCL3dlNode::cbOdom, this);
     }
     if (!params_.fake_imu_)
     {
+      int imu_queue_size;
+      pnh_.param("imu_queue_size", imu_queue_size, 200);
       sub_imu_ = mcl_3dl_compat::subscribe(
           nh_, "imu/data",
-          pnh_, "imu", 200, &MCL3dlNode::cbImu, this);
+          pnh_, "imu", imu_queue_size, &MCL3dlNode::cbImu, this);
     }
 
+    int cloud_queue_size;
+    pnh_.param("cloud_queue_size", cloud_queue_size, 100);
     sub_cloud_ = mcl_3dl_compat::subscribe(
         nh_, "cloud",
-        pnh_, "cloud", 100, &MCL3dlNode::cbCloud, this);
+        pnh_, "cloud", cloud_queue_size, &MCL3dlNode::cbCloud, this);
     sub_mapcloud_ = mcl_3dl_compat::subscribe(
         nh_, "mapcloud",
         pnh_, "mapcloud", 1, &MCL3dlNode::cbMapcloud, this);
@@ -1241,8 +1279,9 @@ public:
     srv_expansion_reset_ = mcl_3dl_compat::advertiseService(
         nh_, "expansion_resetting",
         pnh_, "expansion_resetting", &MCL3dlNode::cbExpansionReset, this);
+    srv_load_pcd_ = nh_.advertiseService("load_pcd", &MCL3dlNode::cbLoadPCD, this);
 
-    point_rep_.setRescaleValues(params_.dist_weight_.data());
+    point_rep_->setRescaleValues(params_.dist_weight_.data());
 
     pf_.reset(new pf::ParticleFilter<State6DOF,
                                      float,
@@ -1308,10 +1347,7 @@ public:
     ROS_DEBUG("max_search_radius: %0.3f", max_search_radius);
     kdtree_.reset(new ChunkedKdtree<PointType>(params_.map_chunk_, max_search_radius));
     kdtree_->setEpsilon(params_.map_grid_min_ / 16);
-    kdtree_->setPointRepresentation(
-        boost::dynamic_pointer_cast<
-            pcl::PointRepresentation<PointType>,
-            MyPointRepresentation>(boost::make_shared<MyPointRepresentation>(point_rep_)));
+    kdtree_->setPointRepresentation(point_rep_);
 
     map_update_timer_ = nh_.createTimer(
         *params_.map_update_interval_,
@@ -1384,6 +1420,7 @@ protected:
   ros::ServiceServer srv_particle_size_;
   ros::ServiceServer srv_global_localization_;
   ros::ServiceServer srv_expansion_reset_;
+  ros::ServiceServer srv_load_pcd_;
 
   tf2_ros::Buffer tfbuf_;
   tf2_ros::TransformListener tfl_;
@@ -1413,7 +1450,7 @@ protected:
   diagnostic_updater::Updater diag_updater_;
   mcl_3dl_msgs::Status status_;
 
-  MyPointRepresentation point_rep_;
+  MyPointRepresentation::Ptr point_rep_;
 
   pcl::PointCloud<PointType>::Ptr pc_map_;
   pcl::PointCloud<PointType>::Ptr pc_map2_;
@@ -1427,7 +1464,8 @@ protected:
 
   std::map<
       std::string,
-      LidarMeasurementModelBase::Ptr> lidar_measurements_;
+      LidarMeasurementModelBase::Ptr>
+      lidar_measurements_;
   ImuMeasurementModelBase::Ptr imu_measurement_model_;
   MotionPredictionModelBase::Ptr motion_prediction_model_;
 
